from ._version import __version__

from tonks.dataloader import MultiDatasetLoader
from tonks.ensemble import *
from tonks.learner import MultiTaskLearner, MultiInputMultiTaskLearner
from tonks.text import *
from tonks.vision import *
